mod camera;
mod renderer;

use std::time::Instant;

use anyhow::Result;
use crawl::winit::event::{Event, WindowEvent};
use renderer::VoxelRenderer;

fn main() -> Result<()> {
    env_logger::init();
    let (mut context, event_loop) = pollster::block_on(
        crawl::ContextBuilder::new()
            .with_title("BEAVER")
            .with_vsync(false)
            .build(),
    )?;

    let mut camera_controller = camera::CameraController::new(
        &context,
        camera::Camera::new(
            glam::Vec3 {
                x: 4.01,
                y: 4.01,
                z: 20.0,
            },
            -90.0_f32.to_radians(),
            0.0_f32.to_radians(),
        ),
        camera::Projection::new(
            context.size.width,
            context.size.height,
            90.0_f32.to_radians(),
            0.01,
            100.0,
        ),
        10.0,
        0.25,
    );

    let renderer = VoxelRenderer::new(&context, &camera_controller)?;
    let passes: Vec<Box<dyn crawl::Pass>> = vec![Box::new(renderer)];

    let mut last_render_time = Instant::now();
    event_loop.run(|event, elwt| match event {
        Event::WindowEvent { window_id, event } if window_id == context.window.id() => {
            if context.handle_window_event(&event, elwt) {
                return;
            }

            if camera_controller.process_events(&event) {
                return;
            }

            if let WindowEvent::RedrawRequested = event {
                let now = Instant::now();
                let dt = now - last_render_time;
                last_render_time = now;

                camera_controller.update(dt);
                camera_controller.update_buffer(&context);

                if let Err(err) = context.render(passes.as_slice()) {
                    log::error!("{err:?}");
                }

                context.window.request_redraw();
            }
        }
        _ => (),
    })?;

    Ok(())
}
