use anyhow::Result;
use crawl::wgpu;

use crate::camera::CameraController;

#[repr(C)]
#[derive(Debug, Default, Copy, Clone, bytemuck::Pod, bytemuck::Zeroable)]
pub struct Ray {
    pos: [f32; 3],
    dist: f32,
    dir: [f32; 3],
    hit: u32,
}

pub struct VoxelRenderer {
    render_texture: crawl::Texture,
    render_pipeline: wgpu::RenderPipeline,
    raygen_bind_group: wgpu::BindGroup,
    raygen_pipeline: wgpu::ComputePipeline,
    raycast_pipeline: wgpu::ComputePipeline,
}

impl VoxelRenderer {
    pub fn new(context: &crawl::Context, camera: &CameraController) -> Result<Self> {
        let render_shader = context
            .device
            .create_shader_module(wgpu::include_wgsl!(concat!(
                env!("CARGO_MANIFEST_DIR"),
                "/res/common/quad.wgsl"
            )));

        let render_texture = crawl::TextureBuilder::new()
            .with_size(context.size.width, context.size.height, 1)
            .with_format(wgpu::TextureFormat::Rgba8Unorm)
            .with_usage(
                wgpu::TextureUsages::TEXTURE_BINDING
                    | wgpu::TextureUsages::COPY_DST
                    | wgpu::TextureUsages::STORAGE_BINDING,
            )
            .with_shader_visibility(wgpu::ShaderStages::FRAGMENT | wgpu::ShaderStages::COMPUTE)
            .build(context)?;

        let render_pipeline = crawl::RenderPipelineBuilder::new("Raycast Quad", &render_shader)
            .with_layout("Draw", &[&render_texture.bind_group_layout], &[])
            .with_fragment_targets(&[Some(context.surface_config.format.into())])
            .build(context);

        let raygen_shader = context
            .device
            .create_shader_module(wgpu::include_wgsl!(concat!(
                env!("CARGO_MANIFEST_DIR"),
                "/res/common/raygen.wgsl"
            )));

        let ray_data = vec![Ray::default(); (context.size.width * context.size.height) as usize];
        let mut raygen_buffers = crawl::BulkBufferBuilder::new()
            .set_usage(wgpu::BufferUsages::STORAGE | wgpu::BufferUsages::COPY_DST)
            .with_init_buffer_bm("Rays", &ray_data)
            .build(context);
        let raygen_layout = crawl::BindGroupLayoutBuilder::new()
            .with_label("Raygen BGL")
            .with_entry(
                wgpu::ShaderStages::COMPUTE,
                wgpu::BindingType::StorageTexture {
                    access: wgpu::StorageTextureAccess::WriteOnly,
                    format: render_texture.attributes.format,
                    view_dimension: wgpu::TextureViewDimension::D2,
                },
                None,
            )
            .with_uniform_entry(wgpu::ShaderStages::COMPUTE)
            .with_rw_storage_entry(wgpu::ShaderStages::COMPUTE)
            .build(context);

        let raygen_bind_group = crawl::BindGroupBuilder::new()
            .with_label("Raygen BG")
            .with_layout(&raygen_layout)
            .with_entry(wgpu::BindingResource::TextureView(&render_texture.view))
            .with_entry(camera.get_buffer().as_entire_binding())
            .with_entry(raygen_buffers.remove(0).as_entire_binding())
            .build(context)?;

        let raygen_pipeline = crawl::ComputePipelineBuilder::new("Raygen P", &raygen_shader)
            .with_layout("Raygen PL", &[&raygen_layout], &[])
            .build(context);

        let raycast_shader = context
            .device
            .create_shader_module(wgpu::include_wgsl!(concat!(
                env!("CARGO_MANIFEST_DIR"),
                "/res/basic_dda/traverse.wgsl"
            )));

        let raycast_pipeline = crawl::ComputePipelineBuilder::new("Raycast P", &raycast_shader)
            .with_layout("Raygen PL", &[&raygen_layout], &[])
            .build(context);

        Ok(Self {
            render_texture,
            render_pipeline,
            raygen_pipeline,
            raygen_bind_group,
            raycast_pipeline,
        })
    }
}

impl crawl::Pass for VoxelRenderer {
    fn execute(&self, encoder: &mut wgpu::CommandEncoder, view: &wgpu::TextureView) {
        let size = self.render_texture.attributes.size;
        let mut compute_pass = encoder.begin_compute_pass(&wgpu::ComputePassDescriptor::default());
        compute_pass.set_pipeline(&self.raygen_pipeline);
        compute_pass.set_bind_group(0, &self.raygen_bind_group, &[]);
        compute_pass.dispatch_workgroups(size.width / 8, size.height / 8, 1);
        drop(compute_pass);

        let mut compute_pass = encoder.begin_compute_pass(&wgpu::ComputePassDescriptor::default());
        compute_pass.set_pipeline(&self.raycast_pipeline);
        compute_pass.set_bind_group(0, &self.raygen_bind_group, &[]);
        compute_pass.dispatch_workgroups(size.width / 8, size.height / 8, 1);
        drop(compute_pass);

        let mut render_pass = encoder.begin_render_pass(&wgpu::RenderPassDescriptor {
            label: Some("Render Pass"),
            color_attachments: &[Some(wgpu::RenderPassColorAttachment {
                view,
                resolve_target: None,
                ops: wgpu::Operations::default(),
            })],
            ..Default::default()
        });
        render_pass.set_pipeline(&self.render_pipeline);
        render_pass.set_bind_group(0, &self.render_texture.bind_group, &[]);
        render_pass.draw(0..6, 0..1);
        drop(render_pass);
    }
}
